from .exceptions import *
from .bc import *
from .spec import instructions


def read_int(s: str) -> int:
    try:
        return int(s)
    except:
        return int(s, 16)


class TASM_Value:
    intc: int

    def __init__(self, type: int, value: int) -> None:
        # Integer
        if type == 10:
            self.intc = 10 + value

        # Bit/Bool
        elif type == 11:
            if value <= 0:
                self.intc = 11
            else:
                self.intc = 12

        # Address
        elif type == 12:
            self.intc = 12 + value

        else:
            raise TASM_UnknownTypeError(type)


class TASM_Instruction:
    def __init__(self, id: int, argval: list[TASM_Value]) -> None:
        self.id = id
        self.argval = argval

    def hexc(self) -> HexCode:
        return intc_to_hexc(self.intc())

    def intc(self) -> IntCode:
        ins: IntCode = [self.id]

        for ii in self.argval:
            ins.append(ii.intc)

        ins.append(0x0)  # end of instruction
        return ins


class Parser:
    src: str
    ins: list[str]

    def __init__(self, src: str, isfile: bool = True) -> None:
        if isfile:
            with open(src, "r") as f:
                self.src = f.read()
        else:
            self.src = src

        # formatting
        self.src = self.src.strip()
        self.ins = self.src.split(";")
        i = 0
        for ii in self.ins:
            self.ins[i] = ii.strip("\n").strip()
            i += 1

    def gen_hc(self) -> HexCode:
        code: HexCode = []
        for instruction in self.ins:
            if instruction == "":
                continue
            parts = instruction.split(" ")
            if parts[0] == "//":
                continue

            elif parts[0] in instructions.keys():
                li = instructions[parts[0]]
                if len(parts) - 1 == li[1]:
                    args = parts[1:]
                    tasm_args: list[TASM_Value] = []
                    argi = 0
                    for ii in args:
                        tasm_args.append(TASM_Value(li[2 + argi], read_int(ii)))
                        argi += 1

                    ins = TASM_Instruction(li[0], tasm_args)
                    for ii in ins.hexc():
                        code.append(ii)

                else:
                    raise TASM_ArgumentError(parts[0], li[1])

            else:
                raise TASM_UnknownInstructionError(parts[0])
        return code



class TVM_Error(Exception):
    """Base class for tvm errors"""

    msg: str

    def __init__(self, msg: str) -> None:
        super().__init__()
        self.msg = msg

    def __str__(self) -> str:
        return self.msg

    def __repr__(self) -> str:
        return str(self)


class TASM_Error(Exception):
    """Base class for tasm errors"""

    msg: str

    def __init__(self, msg: str) -> None:
        super().__init__()
        self.msg = msg

    def __str__(self) -> str:
        return self.msg

    def __repr__(self) -> str:
        return str(self)


class TVM_RegisterIsEmptyError(TVM_Error):
    """
    the register # does not contain a value."""

    def __init__(self, register: int) -> None:
        super().__init__(f"The register {register} does not contain a value.")


class TVM_ValueDecodeError(TVM_Error):
    """Could not decode value."""

    def __init__(self) -> None:
        super().__init__(f"Could not decode value. recompile or update tvm.")


class TVM_UnknownInstructionError(TVM_Error):
    """"""

    def __init__(self, instruction: int) -> None:
        super().__init__(
            f"Unknown instruction '{hex(instruction)}:{str(instruction)}'."
        )


class TVM_UnknownLabelError(TVM_Error):
    def __init__(self, label: int) -> None:
        super().__init__(f"The label '{hex(label)}:{str(label)}' is not defined.")


class TASM_UnknownTypeError(TASM_Error):
    """Unknown type"""

    def __init__(self, type: int) -> None:
        super().__init__(f"Unknown type '{hex(type)}:{type}'.")


class TASM_ArgumentError(TASM_Error):
    """"""

    def __init__(self, instruction: str, argc: int) -> None:
        super().__init__(f"{instruction} requires {str(argc)} arguments.")


class TASM_UnknownInstructionError(TASM_Error):
    """"""

    def __init__(self, instruction: str) -> None:
        super().__init__(f"Unknown instruction '{instruction}'.")


class TASM_IntSizeToBigError(TASM_Error):
    def __init__(self) -> None:
        super().__init__(f"Number to big. Integers must be in range 0-245.")



from .bc import *
from .exceptions import *


def intc_to_val(type: int, val: int) -> int:
    rval = None
    # integer
    if type == 10:
        rval = val - 10

    # bit/bool
    elif type == 11:
        rval = val - 11

    # Address
    elif type == 12:
        rval = val - 12

    else:
        raise TVM_ValueDecodeError()
    return rval


class Register:
    __reg: dict[int]

    def __init__(self) -> None:
        self.__reg = {}

    def set(self, p: int, v):
        try:
            self.__reg[p] = v
        finally:
            print(f"(mem) {hex(p)} <- {v}")

    def get(self, p: int) -> int:
        try:
            return self.__reg[p]
        except KeyError:
            raise TVM_RegisterIsEmptyError(p)
        finally:
            print(f"(mem) {hex(p)} ->")

    def new(self, v=0) -> int:
        na = self.__new_addr()
        self.__reg[na] = v
        return na

    def pop(self, p: int) -> None:
        try:
            self.__reg.pop(p)
        except KeyError:
            raise TVM_RegisterIsEmptyError(p)
        finally:
            print(f"(mem) {hex(p)} <-")

    def __new_addr(self) -> int:
        s = True
        addr = 0
        used = self.__reg.keys()
        while s:
            if addr in used:
                addr += 1
        return addr


class Program:
    register: Register
    running: bool
    retcode: int
    pc: int
    currlabel: int
    labels: dict[int, list[IntCode]]

    def __init__(self, vms: str) -> None:
        self.hc = read_hc_file(vms)
        self.register = Register()

    def eval(self, ic: IntCode) -> None | IntCode:
        try:
            ins_id = ic[0]
            ins_args = ic[1:]

            if ins_id == 10:
                self.ins_mov(*ins_args)

            elif ins_id == 11:
                self.ins_free(*ins_args)

            elif ins_id == 12:
                self.ins_ret(*ins_args)

            elif ins_id == 13:
                self.ins_cpy(*ins_args)

            elif ins_id == 14:
                self.ins_add(*ins_args)

            elif ins_id == 15:
                self.ins_def_label(*ins_args)

            elif ins_id == 16:
                self.ins_jmp(*ins_args)

            else:
                raise TVM_UnknownInstructionError(ins_id)
            return None

        except TVM_Error as e:
            self.retcode = 1
            self.running = False
            print(f"(Error) ({str(self.pc)}) {e}")

        except RecursionError:
            return

    def run_label(self, label: int) -> None:
        try:
            code = self.labels[label]
        except IndexError:
            raise TVM_UnknownLabelError(label)

        for ins in code:
            self.currlabel = label
            self.eval(ins)
        self.currlabel = 0

    def run(self) -> int:
        self.pc = 0
        self.running = True
        self.retcode = 0
        self.currlabel = 0
        self.labels = {0: []}
        self.code = hexc_to_intc(self.hc)
        _ins: IntCode = []

        for _part in self.code:
            if _part == 0x0:
                self.pc += 1
                self.labels[self.currlabel].append(_ins)
                self.eval(_ins)
                _ins = []

                if not self.running:
                    break

            else:
                _ins.append(_part)

        return self.retcode

    def ins_mov(self, _a: int, _b: int) -> None:
        _a = intc_to_val(10, _a)
        _b = intc_to_val(12, _b)
        self.register.set(_b, _a)

    def ins_free(self, _a: int) -> None:
        _a = intc_to_val(12, _a)
        self.register.pop(_a)

    def ins_ret(self, _a: int) -> None:
        _a = intc_to_val(12, _a)
        self.retcode = self.register.get(_a)
        self.running = False

    def ins_cpy(self, _a: int, _b: int) -> None:
        _a = intc_to_val(12, _a)
        _b = intc_to_val(12, _b)
        self.register.set(_b, self.register.get(_a))

    def ins_add(self, _a: int, _b: int) -> None:
        _a = intc_to_val(12, _a)
        _b = intc_to_val(12, _b)
        a = self.register.get(_a)
        b = self.register.get(_b)
        self.register.set(_a, a + b)

    def ins_def_label(self, _a: int) -> None:
        _a = intc_to_val(10, _a)
        self.currlabel = _a
        self.labels[_a] = []

    def ins_jmp(self, _a: int) -> None:
        _a = intc_to_val(10, _a)
        self.run_label(_a)



from typing import TypeAlias
from .exceptions import *


HexCode: TypeAlias = list[str]
IntCode: TypeAlias = list[int]


def intc_to_hexc(ic: IntCode | list[int]) -> HexCode:
    hc = []
    for ii in ic:
        hc.append(hex(ii))
    return hc


def hexc_to_intc(hc: HexCode | list[str]) -> IntCode:
    ic = []
    for ii in hc:
        ic.append(int(ii, 16))
    return ic


def str_to_hexc(s: str) -> HexCode:
    h: HexCode = []
    for ii in s:
        h.append(hex(ord(ii)))

    return h


def hexc_to_str(h: HexCode) -> str:
    s = []
    for ii in h:
        s.append(chr(int(ii, 16)))

    return "".join(s)


def read_hc_file(path: str) -> HexCode:
    rb: bytes = None
    with open(path, "rb") as f:
        rb = f.read()

    rb = list(rb)
    hexcode: HexCode = []
    for dec in rb:
        hexcode.append(hex(dec))

    return hexcode


def write_hc_file(path: str, hexcode: HexCode) -> None:
    try:
        intcode = hexc_to_intc(hexcode)
        bytecode = bytes(intcode)  # to bytes

        with open(path, "wb") as f:  # save
            f.write(bytecode)

    except ValueError:
        raise TASM_IntSizeToBigError()


__all__ = [
    "write_hc_file",
    "read_hc_file",
    "intc_to_hexc",
    "hexc_to_intc",
    "str_to_hexc",
    "hexc_to_str",
    "HexCode",
    "IntCode",
]



import sys

from . import __version__, bc, lang, tvm
from .exceptions import *


def compile(infile: str, outfile: str = "a.tvm") -> None:
    p = lang.Parser(infile)
    hc = p.gen_hc()
    try:
        bc.write_hc_file(outfile, hc)
    except TASM_IntSizeToBigError as e:
        print(f"Error: {e}")


def run(infile: str) -> None:
    p = tvm.Program(infile)
    return p.run()


def usage() -> None:
    print("\nUsage:")
    print("\ttasm help [command]             - Show documentation.")
    print("\ttasm version                    - Shows the current version.")
    print("\ttasm compile <infile> [outfile] - Compiles a tasm program.")
    print("\ttasm run <file>                 - Run a tvm program.")


def help_(command: str) -> None:
    match command:
        case "compile":
            print("Compiles a tasm file to a tvm program.")

        case "run":
            print("Runs a tvm program.")

        case _:
            print("Unknown command.")


def main() -> None:
    argv = sys.argv[1:]

    match argv:
        # compile
        case ["compile", infile]:
            compile(infile)

        case ["compile", infile, ("-o" | "--output"), outfile]:
            compile(infile, outfile)

        # run
        case ["run", infile]:
            return run(infile)

        # help
        case ["help"]:
            print("Compile or run a tasm program with tasm.")
            print("Run 'tasm help <command>' for more documentation.")
            usage()

        case ["help", command]:
            help_(command)

        # version
        case ["version"]:
            print(f"tasm {__version__}")

        #
        case _:
            print("Error: no command given.")
            usage()



from setuptools import setup


setup(
    name="tcc-gh",
    version="1.2.0",
    author="Thijs van Straaten",
    url="https://github.com/ThijsieJWW/tcc",
    packages=["tasm"],
    entry_points={"console_scripts": ["tasm=tasm.cli:main"]},
)
